#include <gst/gst.h>
#include <glib.h>
#include <string.h> /* strcmp() */
#include <unistd.h> /* select() */
#include <signal.h>
#include "input.h"

/* on_pad_added callback data */
struct padlink_cb_data {
    GstElement *aq;     /* audio queue */
    GstElement *vq;     /* video queue */
};

/* mainloop_cb callback data */
struct mainloop_cb_data {
    GMainLoop  *loop;
    GstElement *pipeline;
};

static gint loop_quit = FALSE;

static gboolean bus_call(GstBus     *bus,
                         GstMessage *msg,
                         gpointer    data)
{
    bus = bus;

    GMainLoop *loop = (GMainLoop *) data;

    switch (GST_MESSAGE_TYPE (msg)) {
        case GST_MESSAGE_EOS:
            g_print("End of stream\n");
            g_main_loop_quit(loop);
            break;
        case GST_MESSAGE_ERROR: {
            gchar  *debug;
            GError *error;

            gst_message_parse_error(msg, &error, &debug);
            g_free (debug);

            g_printerr("Error: %s\n", error->message);
            g_error_free(error);

            g_main_loop_quit(loop);
            break;
        }
        default:
            break;
    }
    return TRUE;
}

static void on_pad_added(GstElement *element,
                         GstPad     *pad,
                         gpointer    data)
{
    element = element;

    GstPad     *sinkpad;
    GstElement *decoder;
    gchar      *padname = pad->object.name;

    g_print("avidemux: dynamic pad (%s) created\n", padname);

    if (!strcmp(padname, "video_00"))
        decoder = ((struct padlink_cb_data *) data)->vq;
    else if (!strcmp(padname, "audio_00"))
        decoder = ((struct padlink_cb_data *) data)->aq;
    else
        return;

    sinkpad = gst_element_get_static_pad(decoder, "sink");

    gst_pad_link(pad, sinkpad);
    gst_object_unref(sinkpad);
}

static gboolean loop_idle_cb(gpointer data)
{
    GMainLoop  *loop     = ((struct mainloop_cb_data *)data)->loop;
    GstElement *pipeline = ((struct mainloop_cb_data *)data)->pipeline;
    char key;

    if (loop_quit == TRUE)
        key = 'q';
    else
        key = getch2_read(300);  /* timeout in 0.3 sec */
//    g_print("input: %x(%c)\n", key, key);

    switch (key) {
        case 'p':       /* pause/resume */
        case 0x20: {    /* space bar */
            GstState s;
            if (gst_element_get_state(pipeline, &s, NULL, GST_CLOCK_TIME_NONE)
                    != GST_STATE_CHANGE_SUCCESS)
                break;
            if (s == GST_STATE_PLAYING) {
                g_print("Paused \r");
                s = GST_STATE_PAUSED;
            } else if (s == GST_STATE_PAUSED) {
                g_print("Playing \r");
                s = GST_STATE_PLAYING;
            } else
                break;
            gst_element_set_state(pipeline, s);
            break;
        }
        case 'q':   /* quit */
            g_main_loop_quit(loop);
            break;
    }
    return TRUE;
}

/*
* query information
*/
static gboolean loop_timeout_cb(gpointer data)
{
    GstElement *pipeline = (GstElement *)data;
    GstFormat   fmt = GST_FORMAT_TIME;
    gint64      pos, len;
    GstState    s;

    if ((gst_element_get_state(pipeline, &s, NULL, GST_CLOCK_TIME_NONE)
            != GST_STATE_CHANGE_SUCCESS) || s != GST_STATE_PLAYING)
        return TRUE;

    if (gst_element_query_position(pipeline, &fmt, &pos) &&
        gst_element_query_duration(pipeline, &fmt, &len))
    {
        g_print("Playing - Time: %u:%02u:%02u.%09u / %u:%02u:%02u.%09u\r",
                GST_TIME_ARGS(pos), GST_TIME_ARGS(len));
    }

    return TRUE;
}

static void signalHandler(int signum)
{
    switch (signum) {
    case SIGINT:
        loop_quit = TRUE;
        break;
    case SIGTTOU:
        getch2_setbgflag(TRUE);
        break;
    }
}

int main(int argc, char *argv[])
{
    GMainLoop  *loop;
    GstElement *pipeline, *source, *demuxer, *aqueue, *vqueue;
    GstElement *adecoder, *asink, *vdecoder, *vsink;
    GstBus     *bus;
    gint        ret = 0, ii;

    struct sigaction        sigAction;
    struct padlink_cb_data  pad_cb_data;
    struct mainloop_cb_data idle_cb_data;

    if (argc == 1) {
        g_printerr ("Avi player accepting multiple files from command-line\n");
        g_printerr ("Usage: %s <avi filename> ...\n", argv[0]);
        return -1;
    }

    sigAction.sa_handler = signalHandler;
    sigemptyset(&sigAction.sa_mask);
    sigAction.sa_flags = 0;
    /* insure a clean shutdown if user types ctrl-c */
    sigaction(SIGINT, &sigAction, NULL);
    /* don't read input when runs in background */
    sigaction(SIGTTOU, &sigAction, NULL);
                    
    getch2_enable();
    gst_init(&argc, &argv);

    loop = g_main_loop_new(NULL, FALSE);

    /* Create gstreamer elements */
    pipeline = gst_pipeline_new("aviplayer");
    source   = gst_element_factory_make("filesrc",     "file-source");
    demuxer  = gst_element_factory_make("avidemux",    "avi-demuxer");
    aqueue   = gst_element_factory_make("queue",       "aud-queue");
    vqueue   = gst_element_factory_make("queue",       "vid-queue");
    adecoder = gst_element_factory_make("mad",         "aud-decoder");
    vdecoder = gst_element_factory_make("vdec",        "vid-decoder");
    asink    = gst_element_factory_make("osssink",     "aud-output");
    vsink    = gst_element_factory_make("fbvideosink", "vid-output");

    if (!pipeline || !source   || !demuxer || !aqueue || !vqueue ||
        !adecoder || !vdecoder || !asink   || !vsink) {
        g_printerr("One element could not be created. Exiting.\n");
        ret = -1;
        goto cleanup;
    }

    pad_cb_data.aq = aqueue;
    pad_cb_data.vq = vqueue;

    idle_cb_data.loop     = loop;
    idle_cb_data.pipeline = pipeline;

    /* set component properties */
    g_object_set(G_OBJECT(vdecoder), "engine", "GstEngine", NULL);
    g_object_set(G_OBJECT(vdecoder), "codec", 1, NULL);

    /* add a message handler */
    bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline));
    gst_bus_add_watch(bus, bus_call, loop);
    gst_object_unref(bus);

    /* add all elements into the pipeline */
    gst_bin_add_many(GST_BIN (pipeline), source, demuxer, aqueue,
                     vqueue, adecoder, vdecoder, asink, vsink, NULL);

    /* link the elements together */
    gst_element_link(source, demuxer);
    gst_element_link_many(aqueue, adecoder, asink, NULL);
    gst_element_link_many(vqueue, vdecoder, vsink, NULL);
    g_signal_connect(demuxer, "pad-added", G_CALLBACK(on_pad_added), &pad_cb_data);
    gst_element_set_state(pipeline, GST_STATE_READY);

    for (ii = 1; ii < argc; ii++) {
        g_object_set(G_OBJECT(source), "location", argv[ii], NULL);

        /* Set the pipeline to "playing" state */
        gst_element_set_state(pipeline, GST_STATE_PLAYING);
        g_print("Playing %s...\n", argv[ii]);
        g_idle_add(loop_idle_cb, &idle_cb_data);
        g_timeout_add(300, (GSourceFunc)loop_timeout_cb, pipeline);
        g_main_loop_run(loop);

        /* Out of the main loop, clean up nicely */
        gst_element_set_state(pipeline, GST_STATE_READY);
    }
    gst_element_set_state(pipeline, GST_STATE_NULL);

cleanup:
    gst_object_unref(GST_OBJECT(pipeline));
    g_main_loop_unref(loop);
    gst_deinit();
    getch2_disable();
    g_print("\ndone\n");

    return 0;
}

