/*
 * Plugin Name : fbvideosink
 * Description : A fbdev Video Sink Plugin for TI DM3xx
 *
 *  test pipelines:
 *  gst-launch-0.10 videotestsrc ! video/x-raw-rgb ! fbvideosink device=/dev/fb0
 *  gst-launch-0.10 videotestsrc ! ffmpegcolorspace ! video/x-raw-rgb ! fbvideosink
 *                  device=/dev/fb0
 *  gst-launch-0.10 videotestsrc ! video/x-raw-yuv ! fbvideosink device=/dev/fb3
 *  gst-launch-0.10 videotestsrc ! video/x-raw-yuv,format=\(fourcc\)UYVY ! fbvideosink
 *                  device=/dev/fb3
 *  :: following YUV formats are supported by fbvideosink
 *   : UYVY
 *   : YUY2
 *   : YVYU
 *  :: ffmpegcolorspace supports only YUY2
 *     videotestsrc supports all of the above
 *
 * Copyright (C) 2007 Texas Instruments, Inc.
 * Copyright (C) 2008 Behind The Set, LLC
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation version 2.1 of the License.
 *
 * This program is distributed "as is" WITHOUT ANY WARRANTY of any kind,
 * whether express or implied; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "gstfbsink.h"
#include "gstcbuff.h"
#include <stdlib.h>

//temporary define's
#define ENDIANNESS      G_LITTLE_ENDIAN
#define REDMASK         63488
#define GREENMASK       2016
#define BLUEMASK        31

#include <gst/gstinfo.h>

GST_DEBUG_CATEGORY_STATIC(gst_debug_fbvideosink);
#define GST_CAT_DEFAULT gst_debug_fbvideosink

/* fbimage buffers */
#define GST_TYPE_FBIMAGE_BUFFER (gst_fbimage_buffer_get_type())
#define GST_IS_FBIMAGE_BUFFER(obj) (G_TYPE_CHECK_INSTANCE_TYPE ((obj), \
                                    GST_TYPE_FBIMAGE_BUFFER))
#define GST_FBIMAGE_BUFFER(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), \
                                 GST_TYPE_FBIMAGE_BUFFER, GstFbImageBuffer))

static GstElementDetails gst_fbvideosink_details = {
    "FBDEV Video Sink",
    "Sink/Video",
    "A fbdev based Video Sink",
    "Rishi Bhattacharya <rishi@gstreamer.net>"
};

static GstStaticPadTemplate gst_fbvideosink_sink_template_factory =
    GST_STATIC_PAD_TEMPLATE("sink",
        GST_PAD_SINK,
        GST_PAD_ALWAYS,
        GST_STATIC_CAPS("video/x-raw-yuv, "
                        "framerate = (fraction) [ 0, MAX ], "
                        "width = (int) [ 1, MAX ], "
                        "height = (int) [ 1, MAX ];"
                        "video/x-raw-rgb, "
                        "framerate = (fraction) [ 0, MAX ], "
                        "width = (int) [ 1, MAX ], "
                        "height = (int) [ 1, MAX ]"));

enum {
    PROP_0,
    PROP_DEVICE
};

static gint gst_fbvideosink_get_format_from_caps(GstFbVideoSink *fbsink,
                                                 GstCaps * caps);
static GstVideoSinkClass *parent_class = NULL;

GType gst_fbimage_buffer_get_type(void);
static GstFbContext *gst_fbvideosink_fbcontext_get(GstFbVideoSink * fbsink);
static GstFbImageBuffer *gst_fbvideosink_fbimage_new(GstFbVideoSink * fbsink,
                                                     GstCaps * caps);
static void gst_fbimage_buffer_free(GstFbImageBuffer * fbimage);
static void gst_fbimage_buffer_finalize(GstFbImageBuffer * fbimage);

/* Puts a GstFbImage on a FrameBuffer Display */
static void gst_fbvideosink_fbimage_put(GstFbVideoSink * fbsink,
                                        GstFbImageBuffer * fbimage)
{
    guint offset;
    GstFbContext *fbcb;
    g_return_if_fail(GST_IS_FBVIDEOSINK(fbsink));
    g_return_if_fail(fbsink->fbcb != NULL);

    GST_LOG("BEGIN...");
    fbcb = fbsink->fbcb;
    for (offset = 0; offset < fbcb->nbufs; offset++) {
        if (fbcb->buffers[offset] == GST_BUFFER_DATA(fbimage))
            break;
    }
    if (G_UNLIKELY(offset >= fbcb->nbufs)) {
        GST_ERROR("Invalid framebuffer offset");
        return;
    }
    GST_DEBUG("FB Offset: %d", offset);

    //g_mutex_lock(fbsink->fb_lock);
    fbcb->vinfo.yoffset = fbcb->vinfo.yres * offset;
    if (ioctl(fbcb->fd, FBIOPAN_DISPLAY, &fbcb->vinfo) == -1) {
        //g_mutex_unlock(fbsink->fb_lock);
        GST_ERROR("FBIOPAN_DISPLAY failed");
        return;
    }
    //g_mutex_unlock(fbsink->fb_lock);

    GST_DEBUG("frame ts %llu", GST_BUFFER_TIMESTAMP(fbimage));

    GST_LOG("END..");
}

static GstFlowReturn gst_fbvideosink_show_frame(GstBaseSink * bsink,
                                                GstBuffer * buf)
{
    GstFbVideoSink *fbsink = GST_FBVIDEOSINK(bsink);
    guint8 *tsrc, *tdst;

    GST_LOG("BEGIN..");
//    GST_WARNING("ts: %llu", GST_BUFFER_TIMESTAMP(buf));
    if (GST_IS_FBIMAGE_BUFFER(buf)) {
        GST_DEBUG("fast show");
        gst_fbvideosink_fbimage_put(fbsink, GST_FBIMAGE_BUFFER(buf));
        GST_LOG("END..");
        return GST_FLOW_OK;
    }

    if (!fbsink->fbimage) {
        fbsink->fbimage = gst_fbvideosink_fbimage_new(fbsink,
                               GST_BUFFER_CAPS(buf));
        if (!fbsink->fbimage)
            return GST_FLOW_ERROR;
    }

    if(GST_IS_CONTIGBUFFER(buf) && fbsink->dma != DMACOPY_FAILURE) {
        GST_DEBUG("dma copy");
        tsrc = Memory_getBufferPhysicalAddress(GST_BUFFER_DATA(buf), 4, NULL);
        tdst = Memory_getBufferPhysicalAddress(GST_BUFFER_DATA(fbsink->fbimage), 4, NULL);
        if (Rszcopy_execute(fbsink->dma, tsrc, tdst) == DMACOPY_FAILURE) {
            GST_ERROR("Frame copy using resizer failed");
            fbsink->dma = DMACOPY_FAILURE;
            return GST_FLOW_ERROR;
        }
    }
    else {
        GST_INFO("slow copy");
        tdst = GST_BUFFER_DATA(fbsink->fbimage);
        tsrc = GST_BUFFER_DATA(buf);

        GST_LOG("res %dx%d", fbsink->fbimage->width, fbsink->fbimage->height);
        /* FIXME do memcpy() */
    }

    gst_fbvideosink_fbimage_put(fbsink, fbsink->fbimage);
    GST_LOG("END..");
    return GST_FLOW_OK;
}


static GstFbImageBuffer *gst_fbvideosink_fbimage_new(GstFbVideoSink * fbsink,
                                                     GstCaps * caps)
{
    GstFbImageBuffer *fbimage = NULL;
    GstFbContext *fbcb = fbsink->fbcb;
    static int idx = 0;

    GST_LOG("BEGIN..");
    g_return_val_if_fail(GST_IS_FBVIDEOSINK(fbsink), NULL);

    fbimage = (GstFbImageBuffer *)gst_mini_object_new(GST_TYPE_FBIMAGE_BUFFER);

    fbimage->fbsink = gst_object_ref(fbsink);
    fbimage->size = fbcb->screensize;
    GST_LOG("fbImage Size is %d",fbcb->screensize);

    GST_BUFFER_DATA(fbimage) = fbcb->buffers[idx];
    GST_BUFFER_SIZE(fbimage) = fbimage->size;
    GST_BUFFER_OFFSET(fbimage) = idx;

    GST_DEBUG("Return fbimage %p, offset: %d", GST_BUFFER_DATA(fbimage), idx);
    idx = (idx + 1) % fbcb->nbufs;

    GST_LOG("END..");
    return fbimage;
}

static void gst_fbvideosink_finalize(GObject * object)
{
    GstFbVideoSink *fbsink = GST_FBVIDEOSINK(object);

    GST_LOG("BEGIN..");
    if (fbsink->par) {
        g_free(fbsink->par);
        fbsink->par = NULL;
    }
    if (fbsink->dma != DMACOPY_FAILURE)
        Rszcopy_delete(fbsink->dma);

    G_OBJECT_CLASS(parent_class)->finalize(object);
}

static GstFlowReturn gst_fbvideosink_buffer_alloc(GstBaseSink * bsink,
                                guint64 offset, guint size,
                                GstCaps * caps, GstBuffer ** buf)
{
    GstFbVideoSink *fbsink = GST_FBVIDEOSINK(bsink);

    GST_LOG("BEGIN..");
    *buf = (GstBuffer *)gst_fbvideosink_fbimage_new(fbsink, caps);
    if(*buf == NULL)
        return GST_FLOW_ERROR;

    gst_buffer_set_caps(*buf, caps);
    return GST_FLOW_OK;
}


static void gst_fbimage_buffer_destroy(GstFbImageBuffer * fbimage)
{
    GstFbVideoSink *fbsink = fbimage->fbsink;

    GST_LOG("BEGIN..");
    if (fbsink == NULL)
        goto no_sink;

    g_return_if_fail(GST_IS_FBVIDEOSINK(fbsink));

    /* If the destroyed image is the current one, destroy our reference too */
    if (fbsink->cur_image == fbimage)
        fbsink->cur_image = NULL;

    if (GST_BUFFER_DATA(fbimage)) {
        guint offset = GST_BUFFER_OFFSET(fbimage);

        fbsink->fbcb->avail[offset] = TRUE;
//              g_free ((guint8 *)GST_BUFFER_DATA(fbimage));
        (GST_BUFFER_DATA(fbimage)) = NULL;
        (GST_BUFFER_SIZE(fbimage)) = 0;
    }
    fbimage->fbsink = NULL;
    gst_object_unref(fbsink);

    GST_LOG("END..");
    return;

no_sink:
    GST_WARNING("no sink found");
    return;
}

static void gst_fbimage_buffer_finalize(GstFbImageBuffer * fbimage)
{
    GstFbVideoSink *fbsink = fbimage->fbsink;

    GST_LOG("BEGIN..");
    if (G_UNLIKELY(fbsink == NULL))
        goto no_sink;
#if 0
    /* If our geometry changed we can't reuse that image. */
    if ((fbimage->width != fbsink->video_width) ||
        (fbimage->height != fbsink->video_height)) {
        GST_DEBUG ("destroy fbimage as its size changed %dx%d vs current %dx%d",
                fbimage->width, fbimage->height, fbsink->video_width,
                fbsink->video_height);
        /* we will not destroy it, increment refcount */
//        gst_fbimage_buffer_destroy(fbimage);
        gst_buffer_ref(GST_BUFFER(fbimage));
    }
    else {
#endif
        /* In that case we can reuse the image and add it to our image pool. */
        GST_LOG("recycling fbimage in pool:%x pts:%llu",
                  GST_BUFFER_DATA(fbimage),GST_BUFFER_TIMESTAMP(fbimage));
        /* need to increment the refcount again to recycle */
        gst_buffer_ref(GST_BUFFER(fbimage));
        fbsink->fbcb->avail[GST_BUFFER_OFFSET(fbimage)] = TRUE;
#if 0
    }
#endif
    GST_LOG("END..");
    return;

no_sink:
    GST_WARNING("no sink found");
    return;
}

static void gst_fbimage_buffer_init(GstFbImageBuffer * fbimage,
                                    gpointer g_class)
{
    GST_LOG("BEGIN..");
    GST_LOG("END..");
    return;
}

static void gst_fbimage_buffer_class_init(gpointer g_class,
                                          gpointer class_data)
{
    GstMiniObjectClass *mini_object_class = GST_MINI_OBJECT_CLASS(g_class);

    GST_LOG(".");
    mini_object_class->finalize =
        (GstMiniObjectFinalizeFunction) gst_fbimage_buffer_finalize;
}

GType gst_fbimage_buffer_get_type(void)
{
    static GType _gst_fbimage_buffer_type;

    GST_LOG(".");
    if (G_UNLIKELY(_gst_fbimage_buffer_type == 0)) {
        static const GTypeInfo fbimage_buffer_info = {
            sizeof(GstBufferClass),
            NULL,
            NULL,
            gst_fbimage_buffer_class_init,
            NULL,
            NULL,
            sizeof(GstFbImageBuffer),
            0,
            (GInstanceInitFunc) gst_fbimage_buffer_init,
            NULL
        };
        _gst_fbimage_buffer_type =
            g_type_register_static(GST_TYPE_BUFFER, "GstFbImageBuffer",
                                   &fbimage_buffer_info, 0);
    }
    return _gst_fbimage_buffer_type;
}

/* This function tries to get a format matching with a given caps in the
   supported list of formats we generated in gst_fbvideosink_get_fb_support */
static gint gst_fbvideosink_get_format_from_caps(GstFbVideoSink *fbsink,
                                                 GstCaps * caps)
{
    GList *list = NULL;
    GstCaps *icaps = NULL;

    GST_LOG("BEGIN..");
    g_return_val_if_fail(GST_IS_FBVIDEOSINK(fbsink), 0);

    list = fbsink->fbcb->formats_list;
    while (list) {
        GstFbImageFormat *format = list->data;
        if (format) {
            icaps = gst_caps_intersect(caps, format->caps);
            if (!gst_caps_is_empty(icaps)) {
                gst_caps_unref(icaps);
                GST_LOG("END..");
                return format->format;
            }
        }
        list = g_list_next(list);
    }
    GST_LOG("Return 0 END..");
    return 0;
}

/* This function cleans the Fb context. Closing the Frame Buffer,
    and unrefing the caps for supported formats. */
static void gst_fbvideosink_fbcontext_clear(GstFbVideoSink * fbsink)
{
    GList *formats_list;
    GstFbContext *fbcb;

    GST_LOG("BEGIN..");
    g_return_if_fail(GST_IS_FBVIDEOSINK(fbsink));
    g_return_if_fail(fbsink->fbcb != NULL);

    fbcb = fbsink->fbcb;
    formats_list = fbcb->formats_list;
    while (formats_list) {
        GstFbImageFormat *format = formats_list->data;
        gst_caps_unref(format->caps);
        g_free(format);
        formats_list = g_list_next(formats_list);
    }
    if (fbcb->formats_list)
        g_list_free(fbcb->formats_list);

    gst_caps_unref(fbcb->caps);
    g_free(fbcb->par);

    munmap(fbcb->data, fbcb->screensize * fbcb->nbufs);
    fbcb->data = NULL;
    close(fbcb->fd);
    g_free(fbcb);
    fbcb = NULL;
    GST_LOG("END..");
}

static void gst_fbimage_buffer_free(GstFbImageBuffer * fbimage)
{
    GST_LOG(".");

    /* make sure it is not recycled */
    fbimage->width = -1;
    fbimage->height = -1;
//    gst_buffer_unref(GST_BUFFER(fbimage));
}

/* This function calculates the pixel aspect ratio based on the properties
 * in the fbcb structure and stores it there. */
static void gst_fbvideosink_calculate_pixel_aspect_ratio(GstFbContext * fbcb)
{
    gint par[][2] = {
        {1, 1},                 /* regular screen */
        {16, 15},               /* PAL TV */
        {11, 10},               /* 525 line Rec.601 video */
        {54, 59}                /* 625 line Rec.601 video */
    };
    gint i;
    gint index;
    gdouble ratio;
    gdouble delta;

    GST_LOG("BEGIN..");

#define DELTA(idx) (ABS (ratio - ((gdouble) par[idx][0] / par[idx][1])))

    /* first calculate the "real" ratio based on the FbContext values;
     * which is the "physical" w/h divided by the w/h in pixels of the
     * display */
    ratio = (gdouble) (fbcb->widthmm * fbcb->height) /
            (fbcb->heightmm * fbcb->width);

    GST_DEBUG("calculated pixel aspect ratio: %f", ratio);
    /* now find the one from par[][2] with the lowest delta to the real one */
    delta = DELTA(0);
    index = 0;

    for (i = 1; i < sizeof(par) / (sizeof(gint) * 2); ++i) {
        gdouble this_delta = DELTA(i);

        if (this_delta < delta) {
            index = i;
            delta = this_delta;
        }
    }
    GST_DEBUG("Decided on index %d (%d/%d)", index, par[index][0],
              par[index][1]);

    g_free(fbcb->par);
    fbcb->par = g_new0(GValue, 1);
    g_value_init(fbcb->par, GST_TYPE_FRACTION);
    gst_value_set_fraction(fbcb->par, par[index][0], par[index][1]);
    GST_DEBUG("set fbcb PAR to %d/%d",
              gst_value_get_fraction_numerator(fbcb->par),
              gst_value_get_fraction_denominator(fbcb->par));

    GST_LOG("END..");
}

/* This function gets the fbdev and global info about it. Everything is
   stored in our object and will be cleaned when the object is disposed. Note
   here that caps for supported format are generated without any window or
   image creation */
static GstFbContext *gst_fbvideosink_fbcontext_get(GstFbVideoSink * fbsink)
{
    GstFbContext *fbcb = NULL;
    guint i;

    GST_LOG("BEGIN..");
    g_return_val_if_fail(GST_IS_FBVIDEOSINK(fbsink), NULL);

    /* Allocate 1 element of type GstFbContext and initialize to 0's*/
    fbcb = g_new0(GstFbContext, 1);

    /* Device Driver Open Call for /dev/fb/3 Video Frame Window */
    fbcb->fd = open(fbsink->device, O_RDWR);

    if (fbcb->fd < 0) {
        g_free(fbcb);
        GST_ERROR("Could not open device: %s", fbsink->device);
        return NULL;
    }
    GST_LOG("Display Device Opened");

    // Get fixed screen information
    if (ioctl(fbcb->fd, FBIOGET_FSCREENINFO, &fbcb->finfo) == -1) {
        g_free(fbcb);
        GST_ERROR("Could not get fixed screen info");
        return NULL;
    }

    // Get variable screen information
    if (ioctl(fbcb->fd, FBIOGET_VSCREENINFO, &fbcb->vinfo) == -1) {
        g_free(fbcb);
        GST_ERROR("Could not get variable screen info");
        return NULL;
    }

    fbcb->nbufs = TRIPLE_BUF;
    fbcb->vinfo.yres_virtual = fbcb->vinfo.yres * fbcb->nbufs;
    fbcb->screensize = fbcb->finfo.line_length * fbcb->vinfo.yres;

    /* Set video display format */
    if (ioctl(fbcb->fd, FBIOPUT_VSCREENINFO, &fbcb->vinfo) == -1) {
        g_free(fbcb);
        GST_ERROR("Could not SET variable screen info");
        return NULL;
    }

    fbcb->data = (guint8 *) mmap(0, fbcb->screensize * fbcb->nbufs,
                                 PROT_READ | PROT_WRITE, MAP_SHARED,
                                 fbcb->fd, 0);

    if ((gint) fbcb->data == -1) {
        g_free(fbcb);
        GST_ERROR("failed to map framebuffer device to memory");
        return NULL;
    }

    for (i = 0; i < fbcb->nbufs; i++) {
        fbcb->avail[i] = TRUE;
        fbcb->buffers[i] = fbcb->data + i * fbcb->screensize;
        GST_DEBUG("Mapped Frame Buffer Data Pointer: %p, Offset: %d",
                  fbcb->buffers[i], i);
    }

    gst_fbvideosink_calculate_pixel_aspect_ratio(fbcb);
    if (!fbsink->par) {
        fbsink->par = g_new0(GValue, 1);
        gst_value_init_and_copy(fbsink->par, fbcb->par);
    }

    fbcb->bpp = fbcb->vinfo.bits_per_pixel;
    fbcb->width = fbcb->vinfo.xres;
    fbcb->height = fbcb->vinfo.yres;
    fbcb->widthmm = fbcb->vinfo.width;
    fbcb->heightmm = fbcb->vinfo.height;
    fbcb->endianness = ENDIANNESS;

    GST_DEBUG("Frame Buffer reports %dx%d pixels and %d mm x %d mm",
              fbcb->width, fbcb->height, fbcb->widthmm, fbcb->heightmm);

    fbcb->caps = gst_caps_new_full(
                gst_structure_new("video/x-raw-yuv",
                    "width", GST_TYPE_INT_RANGE, 176, fbcb->width,
                    "height", GST_TYPE_INT_RANGE, 144, fbcb->height,
                    "framerate", GST_TYPE_FRACTION_RANGE, 0, 1, G_MAXINT, 1,
                    "format", GST_TYPE_FOURCC,
                    GST_MAKE_FOURCC('U', 'Y', 'V', 'Y'), NULL),
                gst_structure_new("video/x-raw-yuv",
                    "width", G_TYPE_INT, fbcb->width,
                    "height", G_TYPE_INT, fbcb->height,
                    "framerate", GST_TYPE_FRACTION_RANGE, 0, 1, G_MAXINT, 1,
                    "format", GST_TYPE_FOURCC,
                    GST_MAKE_FOURCC('Y', 'U', 'Y', '2'), NULL),
                gst_structure_new("video/x-raw-yuv",
                    "width", G_TYPE_INT, fbcb->width,
                    "height", G_TYPE_INT, fbcb->height,
                    "framerate", GST_TYPE_FRACTION_RANGE, 0, 1, G_MAXINT, 1,
                    "format", GST_TYPE_FOURCC,
                    GST_MAKE_FOURCC('Y', 'V', 'Y', 'U'), NULL),
                gst_structure_new ("video/x-raw-rgb",
                    "endianness", G_TYPE_INT, fbcb->endianness,
                    "depth", G_TYPE_INT, fbcb->bpp,
                    "bpp", G_TYPE_INT, fbcb->bpp,
                    "blue_mask", G_TYPE_INT, BLUEMASK,
                    "green_mask", G_TYPE_INT, GREENMASK,
                    "red_mask", G_TYPE_INT, REDMASK,
                    "width", G_TYPE_INT, fbcb->width,
                    "height", G_TYPE_INT, fbcb->height,
                    "framerate", GST_TYPE_FRACTION_RANGE, 0, 1, G_MAXINT, 1,
                    NULL), NULL);

    GstFbImageFormat *format = NULL;
    if ((format = g_new0(GstFbImageFormat, 1))) {
        GST_LOG(" I am in UYVY");
        format->format = GST_MAKE_FOURCC('U', 'Y', 'V', 'Y');
        format->caps = gst_caps_new_simple("video/x-raw-yuv",
                        "width", GST_TYPE_INT_RANGE, 176, fbcb->width,
                        "height", GST_TYPE_INT_RANGE, 144, fbcb->height,
                        "framerate", GST_TYPE_FRACTION_RANGE, 0, 1, G_MAXINT, 1,
                        NULL);
        fbcb->formats_list = g_list_append(fbcb->formats_list, format);
    }

    if ((format = g_new0(GstFbImageFormat, 1))) {
        GST_LOG(" I am in YUY2 format");
        format->format = GST_MAKE_FOURCC('Y', 'U', 'Y', '2');
        format->caps = gst_caps_new_simple("video/x-raw-yuv",
                        "width", G_TYPE_INT, fbcb->width,
                        "height", G_TYPE_INT, fbcb->height,
                        "framerate", GST_TYPE_FRACTION_RANGE, 0, 1, G_MAXINT, 1,
                        NULL);
        fbcb->formats_list = g_list_append(fbcb->formats_list, format);
    }

    if ((format = g_new0(GstFbImageFormat, 1))) {
        GST_LOG(" I am in YVYU format");
        format->format = GST_MAKE_FOURCC('Y', 'V', 'Y', 'U');
        format->caps = gst_caps_new_simple("video/x-raw-yuv",
                        "width", G_TYPE_INT, fbcb->width,
                        "height", G_TYPE_INT, fbcb->height,
                        "framerate", GST_TYPE_FRACTION_RANGE, 0, 1, G_MAXINT, 1,
                        NULL);
        fbcb->formats_list = g_list_append(fbcb->formats_list, format);
    }

    if ((format = g_new0(GstFbImageFormat, 1))) {
        GST_LOG(" I am in RGB format");
        format->format = GST_MAKE_FOURCC('R', 'G', 'B', ' ');
        format->caps = gst_caps_new_simple("video/x-raw-rgb",
                        "endianness", G_TYPE_INT, BYTE_ORDER,
                        "depth", G_TYPE_INT, fbcb->bpp,
                        "bpp", G_TYPE_INT, fbcb->bpp,
                        "width", G_TYPE_INT, fbcb->width,
                        "height", G_TYPE_INT, fbcb->height,
                        "framerate", GST_TYPE_FRACTION_RANGE, 0, 1, G_MAXINT, 1,
                        NULL);
        fbcb->formats_list = g_list_append(fbcb->formats_list, format);
    }
    GST_LOG("END..");
    return fbcb;
}

static GstStateChangeReturn gst_fbvideosink_change_state(GstElement * element,
                                                    GstStateChange transition)
{
    GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;
    GstFbVideoSink *fbsink = GST_FBVIDEOSINK(element);

    GST_LOG("BEGIN..");
    switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:
        GST_LOG("change_state from NULL to READY..");
        if (!(fbsink->fbcb = gst_fbvideosink_fbcontext_get(fbsink))) {
            GST_ERROR("Could not create fbcontext");
            return GST_STATE_CHANGE_FAILURE;
        }
//        fbsink->pool_lock = g_mutex_new();
        fbsink->fb_lock = g_mutex_new();
        break;
    case GST_STATE_CHANGE_READY_TO_PAUSED:
        GST_LOG("change_of_state from READY_TO_PAUSED..");
        break;
    case GST_STATE_CHANGE_PAUSED_TO_PLAYING:
        GST_LOG("change_of_state from PAUSED_TO_PLAYING..");
        break;
    default:
        GST_LOG("Invalid Change of state requested..");
        break;
    }

    ret = GST_ELEMENT_CLASS(parent_class)->change_state(element, transition);
    if (ret == GST_STATE_CHANGE_FAILURE) {
        GST_LOG("Return with failure END..");
        return ret;
    }
    switch (transition) {
    case GST_STATE_CHANGE_PLAYING_TO_PAUSED:
        GST_LOG("change_of_state from PLAYING_TO_PAUSED..");
        break;
    case GST_STATE_CHANGE_PAUSED_TO_READY:
        GST_LOG("change_of_state from PAUSED_TO_READY..");
        fbsink->fps_n = 0;
        fbsink->fps_d = 1;
        GST_VIDEO_SINK_WIDTH(fbsink) = 0;
        GST_VIDEO_SINK_HEIGHT(fbsink) = 0;
        break;
    case GST_STATE_CHANGE_READY_TO_NULL:
        GST_LOG("change_of_state from READY_TO_NULL..");
        if (fbsink->fbimage) {
            gst_fbimage_buffer_free(fbsink->fbimage);
            fbsink->fbimage = NULL;
        }
        if (fbsink->pool_lock)
            g_mutex_free(fbsink->pool_lock);
        if (fbsink->fb_lock)
            g_mutex_free(fbsink->fb_lock);
        if (fbsink->fbcb) {
            gst_fbvideosink_fbcontext_clear(fbsink);
            fbsink->fbcb = NULL;
        }
        break;
    default:
        GST_LOG("change_of_state INVALID ..");
        break;
    }
    GST_LOG("END..");
    return ret;
}

static GstCaps *gst_fbvideosink_getcaps(GstBaseSink * bsink)
{
    GstFbVideoSink *fbsink;

    GST_LOG("BEGIN and gst_pad_get_pad_template_caps..");
    fbsink = GST_FBVIDEOSINK(bsink);

    return gst_caps_copy(gst_pad_get_pad_template_caps
                         (GST_VIDEO_SINK_PAD(fbsink)));
}

static gboolean gst_fbvideosink_setcaps(GstBaseSink * bsink,
                                        GstCaps * caps)
{
    GstFbVideoSink *fbsink = GST_FBVIDEOSINK(bsink);
    GstStructure *structure;
    GstCaps *intersection;
    gboolean ret;
    gint video_width, video_height;
    gint video_par_n, video_par_d;      /* video's PAR */
    gint display_par_n, display_par_d;  /* display's PAR */
    GValue display_ratio = { 0, };      /* display w/h ratio */
    const GValue *caps_par;
    const GValue *fps;
    gint num, den;

    GST_DEBUG("BEGIN..");
    GST_DEBUG("Possible caps %p, setting caps %p", fbsink->fbcb->caps, caps);

    intersection = gst_caps_intersect(fbsink->fbcb->caps, caps);
//    GST_DEBUG("intersection returned %" GST_PTR_FORMAT, intersection);

    if (gst_caps_is_empty(intersection)) {
        GST_WARNING("caps intersection empty");
        return FALSE;
    }
    gst_caps_unref(intersection);

    structure = gst_caps_get_structure(caps, 0);
    ret = gst_structure_get_int(structure, "width", &video_width);
    ret &= gst_structure_get_int(structure, "height", &video_height);
    fps = gst_structure_get_value(structure, "framerate");
    ret &= (fps != NULL);

    if (!ret)
        return FALSE;

    fbsink->fps_n = gst_value_get_fraction_numerator(fps);
    fbsink->fps_d = gst_value_get_fraction_denominator(fps);

    if (fbsink->dma != DMACOPY_FAILURE) {
        if (Rszcopy_config(fbsink->dma, video_width, video_height,
                    video_width*2, fbsink->fbcb->finfo.line_length) == DMACOPY_FAILURE ) {
            GST_WARNING("Unable to config resize copy job");
            Rszcopy_delete(fbsink->dma);
            fbsink->dma = DMACOPY_FAILURE;
        }
    }

    GST_LOG("Video %dx%d", video_width, video_height);

    fbsink->video_width = video_width;
    fbsink->video_height = video_height;
    /* get aspect ratio from caps if it's present, and
     * convert video width and height to a display width and height
     * using wd / hd = wv / hv * PARv / PARd
     * the ratio wd / hd will be stored in display_ratio */
    g_value_init(&display_ratio, GST_TYPE_FRACTION);

    /* get video's PAR */
    caps_par = gst_structure_get_value(structure, "pixel-aspect-ratio");
    GST_LOG("Pixel Aspect Ratio %d",caps_par);

    if (caps_par) {
        video_par_n = gst_value_get_fraction_numerator(caps_par);
        video_par_d = gst_value_get_fraction_denominator(caps_par);
    }
    else {
        video_par_n = 1;
        video_par_d = 1;
    }

    /* get display's PAR */
    if (fbsink->par) {
        display_par_n = gst_value_get_fraction_numerator(fbsink->par);
        display_par_d = gst_value_get_fraction_denominator(fbsink->par);
    }
    else {
        display_par_n = 1;
        display_par_d = 1;
    }

    GST_LOG("video par n: %d, display par n: %d, "
            "video par d: %d, display par d: %d",
            video_par_n,display_par_n, video_par_d,display_par_d);

    gst_value_set_fraction(&display_ratio,
                           video_width * video_par_n * display_par_d,
                           video_height * video_par_d * display_par_n);

    num = gst_value_get_fraction_numerator(&display_ratio);
    den = gst_value_get_fraction_denominator(&display_ratio);
    GST_LOG("video width/height: %dx%d, calculated display ratio: %d/%d",
            video_width, video_height, num, den);

    /* now find a width x height that respects this display ratio.
     * prefer those that have one of w/h the same as the incoming video
     * using wd / hd = num / den
     */

    /* start with same height, because of interlaced video */
    /* check hd / den is an integer scale factor, and scale wd with the PAR */
    if (video_height % den == 0) {
        GST_DEBUG("keeping video height");
        GST_VIDEO_SINK_WIDTH(fbsink) = video_height * num / den;
        GST_VIDEO_SINK_HEIGHT(fbsink) = video_height;
    }
    else if (video_width % num == 0) {
        GST_DEBUG("keeping video width");
        GST_VIDEO_SINK_WIDTH(fbsink) = video_width;
        GST_VIDEO_SINK_HEIGHT(fbsink) = video_width * den / num;
    }
    else {
        GST_DEBUG( "approximating while keeping video height");
        GST_VIDEO_SINK_WIDTH(fbsink) = video_height * num / den;
        GST_VIDEO_SINK_HEIGHT(fbsink) = video_height;
    }
#if 0
    /* We renew our fbimage only if size or format changed;
     * the fbimage is the same size as the video pixel size */
    if (fbsink->fbimage &&
        ((video_width != fbsink->fbimage->width) ||
         (video_height != fbsink->fbimage->height))) {
        gst_buffer_unref(GST_BUFFER(fbsink->fbimage));
        fbsink->fbimage = NULL;
    }
#endif
    GST_LOG("END..");
    return TRUE;
}

/*...  Init and Class Init ...*/

static void gst_fbvideosink_set_property(GObject * object, guint prop_id,
                                     const GValue * value, GParamSpec * pspec)
{
    GstFbVideoSink *fbsink = GST_FBVIDEOSINK(object);

    GST_LOG(".");
    switch (prop_id) {
    case PROP_DEVICE:
        GST_LOG("Set_property PROP_DEVICE..");
        g_free(fbsink->device);
        fbsink->device = g_value_dup_string(value);
        break;
    default:
        GST_LOG("Invalid property ID..");
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void gst_fbvideosink_get_property(GObject * object, guint prop_id,
                                         GValue * value, GParamSpec * pspec)
{
    GstFbVideoSink *fbsink;

    GST_LOG(".");
    fbsink = GST_FBVIDEOSINK(object);
    switch (prop_id) {
    case PROP_DEVICE:
        g_value_set_string(value, fbsink->device);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

gboolean gst_fbvideosink_query(GstElement * element, GstQuery * query)
{
    GstStructure *vidp;
    GValue framesDisplayed = { 0, };
    GstFbVideoSink *fbsink = GST_FBVIDEOSINK(element);

    GST_LOG("BEGIN..");
    g_value_init(&framesDisplayed, G_TYPE_UINT);
    g_value_set_uint(&framesDisplayed, fbsink->framesDisplayed);

    vidp = gst_query_get_structure(query);
    gst_structure_set_value(vidp, "framesDisplayed", &framesDisplayed);
    g_value_unset(&framesDisplayed);

    GST_LOG("END..");
    return TRUE;
}

static void gst_fbvideosink_init(GstFbVideoSink * fbsink)
{
    GstElementClass *element_class =
        GST_ELEMENT_CLASS(GST_FBVIDEOSINK_GET_CLASS(fbsink));

    GST_LOG(".");

    fbsink->fbcb = NULL;
    fbsink->fbimage = NULL;
    fbsink->cur_image = NULL;
    fbsink->framesDisplayed = 0;

    fbsink->fps_n = 0;
    fbsink->fps_d = 0;
    fbsink->video_width = 0;
    fbsink->video_height = 0;

    fbsink->par = NULL;
    fbsink->device = g_strdup(DEFAULT_DEVICE);
    element_class->query = gst_fbvideosink_query;

    fbsink->dma = Rszcopy_create(DMACOPY_DEFAULTRSZRATE);
    if (fbsink->dma == DMACOPY_FAILURE) {
        GST_WARNING("Failed to create resize copy job");
    }
}

static void gst_fbvideosink_base_init(gpointer g_class)
{
    GstElementClass *element_class = GST_ELEMENT_CLASS(g_class);

    GST_LOG(".");
    gst_element_class_set_details(element_class, &gst_fbvideosink_details);
    gst_element_class_add_pad_template(element_class,
        gst_static_pad_template_get(&gst_fbvideosink_sink_template_factory));
}

static void gst_fbvideosink_class_init(GstFbVideoSinkClass * klass)
{
    GObjectClass *gobject_class;
    GstElementClass *gstelement_class;
    GstBaseSinkClass *gstbasesink_class;

    GST_DEBUG_CATEGORY_INIT(gst_debug_fbvideosink, "fbvideosink", 0,
                            "fbvideosink element");

    GST_LOG(".");

    gobject_class = (GObjectClass *) klass;
    gstelement_class = (GstElementClass *) klass;
    gstbasesink_class = (GstBaseSinkClass *) klass;

    parent_class = g_type_class_ref(GST_TYPE_VIDEO_SINK);

    gobject_class->set_property = gst_fbvideosink_set_property;
    gobject_class->get_property = gst_fbvideosink_get_property;
    g_object_class_install_property(gobject_class, PROP_DEVICE,
            g_param_spec_string("device", "Device",
                                "Frame Buffer Device (usually /dev/fb0)",
                                DEFAULT_DEVICE, G_PARAM_READWRITE));

    gobject_class->finalize = gst_fbvideosink_finalize;
    gstelement_class->change_state =
        GST_DEBUG_FUNCPTR(gst_fbvideosink_change_state);

    gstbasesink_class->get_caps =
        GST_DEBUG_FUNCPTR(gst_fbvideosink_getcaps);
    gstbasesink_class->set_caps =
        GST_DEBUG_FUNCPTR(gst_fbvideosink_setcaps);

    gstbasesink_class->buffer_alloc =
        GST_DEBUG_FUNCPTR(gst_fbvideosink_buffer_alloc);
    gstbasesink_class->preroll =
        GST_DEBUG_FUNCPTR(gst_fbvideosink_show_frame);
    gstbasesink_class->render =
        GST_DEBUG_FUNCPTR(gst_fbvideosink_show_frame);
}

GType gst_fbvideosink_get_type(void)
{
    static GType fbvideosink_type = 0;

    GST_LOG(".");
    if (!fbvideosink_type) {
        static const GTypeInfo fbvideosink_info = {
            sizeof(GstFbVideoSinkClass),
            gst_fbvideosink_base_init,
            NULL,
            (GClassInitFunc) gst_fbvideosink_class_init,
            NULL,
            NULL,
            sizeof(GstFbVideoSink),
            0,
            (GInstanceInitFunc) gst_fbvideosink_init,
        };

        fbvideosink_type =
            g_type_register_static(GST_TYPE_VIDEO_SINK, "GstFbVideoSink",
                                   &fbvideosink_info, 0);
    }
    return fbvideosink_type;
}
