/*
 * Plugin Name : vdec
 * Description : A Generic (ARM based) Video Decoder for for TI DM3xx
 *
 * Copyright (C) 2008 Texas Instruments, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation version 2.1 of the License.
 *
 * This program is distributed "as is" WITHOUT ANY WARRANTY of any kind,
 * whether express or implied; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "gstvdec.h"
#include "gstfbsink.h"     /* for plugin_init */

GST_DEBUG_CATEGORY_STATIC(gstvdec_debug);
#define GST_CAT_DEFAULT (gstvdec_debug)

typedef enum {
    ST_VIDEO_H264,
    ST_VIDEO_MPEG4,
    ST_VIDEO_WMV3,
    ST_VIDEO_WVC1
} CODECS;

/* Properties Type Enumeration */
typedef enum {
    PROP_0,                 //prop_id should be greater than 0.
    ARG_ENGINE_NAME,
    ARG_ENGINE_OPEN,
    ARG_CODEC_NAME,
    ARG_PAD_BUF_ALLOC
} PROPERTIES;

/* Engine Type Enumeration */
typedef enum {
    H264ENGINE,
    MPEG4ENGINE,
} ENGINE;

/* Declarations */
static void gst_vdec_base_init(GstVDecClass * klass);
static void gst_vdec_class_init(GstVDecClass * klass);
static void gst_vdec_init(GstVDec * vdec);
static GstFlowReturn gst_vdec_chain(GstPad * pad,GstBuffer * buffer);
static GstStateChangeReturn gst_vdec_change_state(GstElement * element,
                                                  GstStateChange trans);

static gboolean gst_vdec_codec_init(GstVDec * vdec);
static GstCaps *gst_vdec_sink_getcaps(GstPad * pad);
static gboolean gst_vdec_sink_setcaps(GstPad * pad, GstCaps * caps);

static gboolean gst_vdec_src_event(GstPad * pad, GstEvent * event);
static gboolean gst_vdec_sink_event(GstPad * pad, GstEvent * event);

static gboolean gst_vdec_sink_activate(GstPad * sinkpad);
static gboolean gst_vdec_sink_activate_pull(GstPad * sinkpad, gboolean active);
static void gst_vdec_loop(GstPad * pad);
static void gst_vdec_set_property(GObject * object, guint prop_id,
                                  const GValue * value, GParamSpec * pspec);


/* elementfactory information */
static GstElementDetails vdec_details = {
    "A Generic Video Decoder",
    "Codec/Decoder/Video",
    "Decodes Various Video Formats using the Codec Engine",
    "Rishi Bhattacharya <rishi@gstreamer.net>"
    "                Pratheesh Gangadhar <pratheesh@ti.com>"
};

static GstStaticPadTemplate sink_template = GST_STATIC_PAD_TEMPLATE("sink",
        GST_PAD_SINK,
        GST_PAD_ALWAYS,
        GST_STATIC_CAPS
        ("video/mpeg, "
         "mpegversion = (int) { 4}, "
         "systemstream = (boolean) FALSE; "
         "video/x-xvid;"
         "video/x-h264"));

static GstStaticPadTemplate src_template = GST_STATIC_PAD_TEMPLATE("src",
        GST_PAD_SRC,
        GST_PAD_ALWAYS,
        GST_STATIC_CAPS
        ("video/x-raw-yuv, "
         "format = (fourcc) UYVY , "
         "framerate = (fraction) [ 0, MAX ], "
         "width = (int) [ 1, MAX ], "
         "height = (int) [ 1, MAX ];"
         "video/x-raw-yuv, "
         "format = (fourcc) YUY2 , "
         "framerate = (fraction) [ 0, MAX ], "
         "width = (int) [ 1, MAX ], "
         "height = (int) [ 1, MAX ];"
         "video/x-raw-yuv, "
         "format = (fourcc) YVYU , "
         "framerate = (fraction) [ 0, MAX ], "
         "width = (int) [ 1, MAX ], "
         "height = (int) [ 1, MAX ]"));


static GstElementClass *parent_class = NULL;

static GType gst_vdec_get_codec_type(void)
{
    static GType vdec_codec_type = 0;

    GST_LOG("Begin");
    if (!vdec_codec_type) {
        static GEnumValue codec_types[] = {
            {ST_VIDEO_H264, "H264", "H264 raw bit stream"},
            {ST_VIDEO_MPEG4, "MPEG4", "MPEG4  raw bit stream"},
            {0, NULL, NULL}
        };
        vdec_codec_type =
            g_enum_register_static("VDecCodecName", codec_types);
    }
    return vdec_codec_type;
}


GType gst_vdec_get_type(void)
{
    static GType vdec_type = 0;

    if (!vdec_type) {
        static const GTypeInfo vdec_info = {
            sizeof(GstVDecClass),
            (GBaseInitFunc) gst_vdec_base_init,
            NULL,                   /* base_finalize function is NULL */
            (GClassInitFunc) gst_vdec_class_init,
            NULL,                   /* class_finalize is NULL */
            NULL,                   /* class_data is NULL */
            sizeof(GstVDec),        /* Instance type */
            0,                      /* Number of pre-allocated Instances */
            (GInstanceInitFunc) gst_vdec_init,
        };
        vdec_type =
            g_type_register_static(GST_TYPE_ELEMENT, "GstVDec", &vdec_info, 0);

        GST_DEBUG_CATEGORY_INIT(gstvdec_debug, "vdec", 0,
                                "Generic Video Decoder Element for the ARM");
    }
    return vdec_type;
}

static GstCaps *gst_vdec_sink_getcaps(GstPad * pad)
{
    GstVDec *vdec;

    GST_LOG("invoked and call gst_caps_copy");
    vdec = GST_VDEC(GST_PAD_PARENT(pad));
    return gst_caps_copy(gst_pad_get_pad_template_caps(pad));
}

static gboolean gst_vdec_sink_setcaps(GstPad * pad, GstCaps * caps)
{
    GstVDec *vdec = GST_VDEC(GST_PAD_PARENT(pad));
    GstStructure *structure;
    GstCaps *intersection;
    const gchar *mime;

    GST_LOG("BEGIN...");

    intersection =
        gst_caps_intersect(gst_pad_get_pad_template_caps(pad), caps);
//    GST_DEBUG_OBJECT(vdec, "Intersection Returned %" GST_PTR_FORMAT,
//                     intersection);

    if (gst_caps_is_empty(intersection)) {
        GST_ERROR("Caps empty...");
        return FALSE;
    }
    gst_caps_unref(intersection);

    structure = gst_caps_get_structure(caps, 0);
    mime = gst_structure_get_name(structure);

    if (!strcmp("video/x-h264", mime)) {
        vdec->codec = ST_VIDEO_H264;
        GST_INFO("H.264");
    }
    else if (!strcmp("video/mpeg", mime)) {
        int version = 0;
        gst_structure_get_int(structure, "mpegversion", &version);

        if (version == 4) {
            vdec->codec = ST_VIDEO_MPEG4;
            GST_INFO("MPEG4");
        }
        else
            return FALSE;
    }
    else if (!strcmp("video/x-xvid", mime)) {
        vdec->codec = ST_VIDEO_MPEG4;
        GST_DEBUG("XVID");
    }
    else {
        GST_WARNING("Unknown mime type...");
        return FALSE;
    }
    vdec->isCodecPropSet = TRUE;

    gst_structure_get_int(structure, "width", &vdec->width);
    gst_structure_get_int(structure, "height", &vdec->height);
    return TRUE;
}

static void gst_vdec_base_init(GstVDecClass * klass)
{
    GstElementClass *element_class = GST_ELEMENT_CLASS(klass);

    GST_LOG("BEGIN...");

    klass->sink_template = gst_static_pad_template_get(&sink_template);
    klass->src_template = gst_static_pad_template_get(&src_template);

    gst_element_class_add_pad_template(element_class, klass->src_template);
    gst_element_class_add_pad_template(element_class,
                                       klass->sink_template);

    gst_element_class_set_details(element_class, &vdec_details);
}

static void gst_vdec_class_init(GstVDecClass * klass)
{
    GstElementClass *gstelement_class;
    GObjectClass *object_class = (GObjectClass *) (klass);

    GST_LOG("BEGIN...");

    parent_class = g_type_class_ref(GST_TYPE_ELEMENT);
    gstelement_class = (GstElementClass *) klass;
    gstelement_class->change_state = gst_vdec_change_state;
    object_class->set_property     = gst_vdec_set_property;

    g_object_class_install_property(object_class, ARG_ENGINE_OPEN,
            g_param_spec_string("engine", "Engine",
                                "Force a specific engine to use",
                                "", G_PARAM_WRITABLE));

    g_object_class_install_property(object_class, ARG_CODEC_NAME,
            g_param_spec_enum("codec", "Codec",
                              "Name of Codec to run",
                              GST_TYPE_VDEC_CODEC,
                              0,
                              G_PARAM_WRITABLE));

    g_object_class_install_property(object_class, ARG_PAD_BUF_ALLOC,
            g_param_spec_boolean("pad-buf-alloc", "PadBufAlloc",
                              "Whether to use pad_buffer_alloc from downstram element",
                              TRUE, G_PARAM_WRITABLE));
}

static void gst_vdec_set_property(GObject * object, guint prop_id,
                                  const GValue * value,
                                  GParamSpec * pspec)
{
    GstVDec *vdec = GST_VDEC(object);

    GST_LOG("BEGIN...%d..",prop_id);
    switch (prop_id) {
    case ARG_ENGINE_OPEN:
        if(vdec->engineName)
            free(vdec->engineName);
        vdec->engineName = g_value_dup_string(value);
        GST_LOG("Engine name set to value %s", vdec->engineName);
        break;

    case ARG_CODEC_NAME:
        vdec->codec = g_value_get_enum(value);
        vdec->isCodecPropSet = TRUE;
        GST_LOG("Codec set to enum value %d", vdec->codec);
        break;

    case ARG_PAD_BUF_ALLOC:
        vdec->padBufAlloc = g_value_get_boolean(value);
        GST_LOG("padBufAlloc set to boolean value %d", vdec->padBufAlloc);

    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

gboolean gst_vdec_query(GstElement * element, GstQuery * query)
{
    GstStructure *vidp;
    GValue bytesConsumed = { 0, };
    GValue dspLoad       = { 0, };
    GstVDec *vdec        = GST_VDEC(element);

    GST_LOG("BEGIN..");
    g_value_init(&bytesConsumed, G_TYPE_UINT64);
//    g_value_set_uint64(&bytesConsumed, vdec->bytesConsumed);

    g_value_init(&dspLoad, G_TYPE_INT);
    g_value_set_int(&dspLoad, Engine_getCpuLoad(vdec->ce));

    vidp = gst_query_get_structure(query);
    gst_structure_set_value(vidp, "bytesConsumed", &bytesConsumed);
    gst_structure_set_value(vidp, "dspCpuLoad", &dspLoad);
    g_value_unset(&bytesConsumed);
    g_value_unset(&dspLoad);

    return TRUE;
}


static void gst_vdec_init(GstVDec * vdec)
{
    GstVDecClass *klass = GST_VDEC_GET_CLASS(vdec);
    GstElementClass *element_class = GST_ELEMENT_CLASS(klass);

    GST_LOG("BEGIN..");
    vdec->sinkpad = gst_pad_new_from_template(klass->sink_template, "sink");
    gst_pad_set_activate_function(vdec->sinkpad,
                            GST_DEBUG_FUNCPTR (gst_vdec_sink_activate));
    gst_pad_set_activatepull_function(vdec->sinkpad,
                            GST_DEBUG_FUNCPTR (gst_vdec_sink_activate_pull));

    gst_element_add_pad(GST_ELEMENT(vdec), vdec->sinkpad);
    gst_pad_set_event_function(vdec->sinkpad,
                            GST_DEBUG_FUNCPTR(gst_vdec_sink_event));

    gst_pad_set_chain_function(vdec->sinkpad,
                            GST_DEBUG_FUNCPTR(gst_vdec_chain));
    gst_pad_set_getcaps_function(vdec->sinkpad, gst_vdec_sink_getcaps);
    gst_pad_set_setcaps_function(vdec->sinkpad, gst_vdec_sink_setcaps);

    vdec->srcpad = gst_pad_new_from_template(klass->src_template, "src");
    gst_element_add_pad(GST_ELEMENT(vdec), vdec->srcpad);
    gst_pad_set_event_function(vdec->srcpad,
                            GST_DEBUG_FUNCPTR(gst_vdec_src_event));
    element_class->query = gst_vdec_query;
}


static gboolean gst_vdec_sink_activate(GstPad * sinkpad)
{
    GstVDec *vdec = GST_VDEC(gst_pad_get_parent(sinkpad));

    GST_LOG("Begin");
    if (gst_pad_check_pull_range(sinkpad)) {
        vdec->streaming = FALSE;
        GST_DEBUG("STREAMING is FALSE, Pull mode is not supported");
        gst_object_unref(vdec);
        return gst_pad_activate_pull(sinkpad, TRUE);
    } else {
        vdec->streaming = TRUE;
        GST_DEBUG("STREAMING is TRUE");
        gst_object_unref(vdec);
        return gst_pad_activate_push(sinkpad, TRUE);
    }
}

static gboolean
gst_vdec_sink_activate_pull(GstPad * sinkpad, gboolean active)
{
    GstVDec *vdec = GST_VDEC(gst_pad_get_parent(sinkpad));

    GST_LOG("BEGIN..");
    if (active) {
        /* if we have a scheduler we can start the task */
        gst_pad_start_task(sinkpad, (GstTaskFunction)gst_vdec_loop, sinkpad);
    } else {
        gst_pad_stop_task(sinkpad);
    }
    gst_object_unref(vdec);
    return TRUE;
}

/* gst_vdec_decode accepts output buffer to fill as argument and updates it
   with decoded output and time stamp to display. If no output  is avaiable,
   will set GST_CLOCK_TIME_NONE as time stamp
*/

static gint gst_vdec_decode(GstVDec * vdec, GstBuffer ** outbuf)
{
    IVIDDEC2_InArgs in_arg;
    IVIDDEC2_OutArgs out_arg;

    XDAS_Int32 inBufSize, outBufSize, ret;
    XDAS_Int8 *outData, *inData;
    guint64 timestamp;
    guint64 duration;

    GST_LOG("BEGIN..");

    inBufSize = GST_BUFFER_SIZE(vdec->sinkbuf);
    timestamp = GST_BUFFER_TIMESTAMP(vdec->sinkbuf);
    duration  = GST_BUFFER_DURATION(vdec->sinkbuf);

    inData = (XDAS_Int8 *) GST_BUFFER_DATA(vdec->sinkbuf);
    GST_LOG("inBufSize = %d, at 0x%p of 0x%p",
            inBufSize, inData, vdec->sinkbuf);

//    GST_DEBUG("vdec: inbuf size = %d , ptr = %p", inBufSize, inData);

    //Fill in codec engine input parameters before process call
    vdec->inBufDesc.descs[0].bufSize = inBufSize;
    vdec->inBufDesc.numBufs          = 1;
    vdec->inBufDesc.descs[0].buf     = inData;

    outBufSize = GST_BUFFER_SIZE(*outbuf);
    outData    = (XDAS_Int8 *) GST_BUFFER_DATA(*outbuf);

    GST_LOG("outbuf size = %d, at %p", outBufSize, outData);

    vdec->outBufDesc.bufSizes = &outBufSize;
    vdec->outBufDesc.bufs     = &outData;
    vdec->outBufDesc.numBufs  = 1;

    GST_BUFFER_TIMESTAMP(*outbuf) = timestamp;
    GST_BUFFER_DURATION(*outbuf)  = duration;
    in_arg.inputID                = (XDAS_Int32) *outbuf;

    in_arg.size     = sizeof(IVIDDEC2_InArgs);
    in_arg.numBytes = inBufSize;
    out_arg.size    = sizeof(IVIDDEC2_OutArgs);
//    out_arg.outputID = 0; /*FIXME check how codec uses outputID */

    ret = VIDDEC2_process((VIDDEC2_Handle) vdec->decoder,
                          &vdec->inBufDesc, &vdec->outBufDesc,
                          &in_arg, &out_arg);
    GST_LOG("decoded %d bytes", out_arg.bytesConsumed);

    if (ret != VIDDEC2_EOK) {
        GST_WARNING("VIDDEC2_process() failed, ret: %d, error: 0x%x", 
                    ret, out_arg.decodedBufs.extendedError);
        return -1;
    }

    //If buffer can't be displayed, set time stamp to GST_CLOCK_TIME_NONE
    if (out_arg.displayBufs[0].numBufs == 0) {
        GST_WARNING("vidDecOutArlayBufs.numBufs == 0");

        GST_BUFFER_TIMESTAMP(*outbuf) = GST_CLOCK_TIME_NONE;
    } else {
//            *outbuf = (GstBuffer *)out_arg.outputID;
        GST_BUFFER_TIMESTAMP(outbuf) = timestamp;
        GST_BUFFER_SIZE(*outbuf)     = out_arg.displayBufs[0].bufDesc[0].bufSize;
    }
    GST_LOG("END..");
    return 0;
}


static gint gst_vdec_process(GstVDec * vdec, GstBuffer * inbuf)
{
    GstCaps *othercaps;

    GST_LOG("BEGIN");
    if (vdec->sinkbuf == NULL) {
        vdec->sinkbuf = gst_contigbuffer_new(vdec->inBufSize);
        if (vdec->sinkbuf == NULL) {
            GST_WARNING("Failed to allocate sinkbuf");
            return GST_FLOW_UNEXPECTED;
        }
    }
    memcpy(GST_BUFFER_DATA(vdec->sinkbuf), GST_BUFFER_DATA(inbuf),
           MIN(vdec->inBufSize, GST_BUFFER_SIZE(inbuf)));

    GST_BUFFER_TIMESTAMP (vdec->sinkbuf) = GST_BUFFER_TIMESTAMP(inbuf);
    GST_BUFFER_SIZE (vdec->sinkbuf)      = GST_BUFFER_SIZE(inbuf);
    GST_BUFFER_DURATION (vdec->sinkbuf)  = GST_BUFFER_DURATION(inbuf);
    gst_buffer_set_caps (vdec->sinkbuf, GST_PAD_CAPS (vdec->sinkpad));
    gst_buffer_unref(inbuf);

    if (vdec->padBufAlloc || vdec->firstFrame)
        gst_pad_alloc_buffer(vdec->srcpad, GST_BUFFER_OFFSET_NONE,
                     vdec->width * vdec->height * 2,
                     GST_PAD_CAPS(vdec->srcpad), &vdec->srcbuf);
    else
        vdec->srcbuf = gst_contigbuffer_new(vdec->width * vdec->height * 2);

    if (vdec->srcbuf == NULL) {
        GST_WARNING("Failed to allocate outbuf");
        return -2;
    }

    if (gst_vdec_decode(vdec, &vdec->srcbuf) < 0) {
        GST_WARNING("decode failed");
    }
    else if (GST_BUFFER_TIMESTAMP(vdec->srcbuf) != GST_CLOCK_TIME_NONE) {
//        GST_WARNING("ts: %llu", GST_BUFFER_TIMESTAMP(vdec->srcbuf));
        gst_buffer_set_caps(vdec->srcbuf, GST_PAD_CAPS(vdec->srcpad));
        if(gst_pad_push(vdec->srcpad, vdec->srcbuf) != GST_FLOW_OK) {
            GST_DEBUG("pad push failed, ts: %llu",
                GST_BUFFER_TIMESTAMP(vdec->srcbuf));
            return GST_FLOW_UNEXPECTED;
        }
    }

    if (vdec->firstFrame) {
        gst_segment_init(&vdec->segment, GST_FORMAT_TIME);
        othercaps = gst_caps_new_simple("video/x-raw-yuv", "format",
                    GST_TYPE_FOURCC, GST_MAKE_FOURCC('U', 'Y', 'V', 'Y'),
                    "framerate", GST_TYPE_FRACTION, 30000, 1001,
                    "width", G_TYPE_INT, vdec->width,
                    "height", G_TYPE_INT, vdec->height, NULL);

        gst_pad_set_caps(vdec->srcpad, othercaps);
        gst_pad_use_fixed_caps(vdec->srcpad);
        gst_caps_unref(othercaps);

        vdec->firstFrame = FALSE;
    }
    return GST_FLOW_OK;
}

/*TODO : Not supported now, process is not generic enough to handle this*/
static void gst_vdec_loop(GstPad * pad)
{
#if 0
    XDAS_Int32 ret;
    GstVDec *vdec = GST_VDEC(gst_pad_get_parent(pad));

    GST_LOG("BEGIN..");
    if ((ret = gst_vdec_process(vdec)) == -2)
        gst_pad_push_event(vdec->srcpad, gst_event_new_eos());
#endif
}

static GstFlowReturn gst_vdec_chain(GstPad * pad, GstBuffer * buffer)
{
    GstVDec *vdec = GST_VDEC(gst_pad_get_parent(pad));
    gint ret = GST_FLOW_OK;

    GST_LOG("received buffer ts %llu, size %d at 0x%p of 0x%p",
                GST_BUFFER_TIMESTAMP(buffer), GST_BUFFER_SIZE(buffer),
                GST_BUFFER_DATA(buffer), buffer);

    if (gst_vdec_process(vdec, buffer) == -2) {
        gst_pad_push_event(vdec->srcpad, gst_event_new_eos());
        ret = GST_FLOW_ERROR;
    }
    gst_object_unref(vdec);
    return ret;
}

/* open codec engine and create a decoder instance */
static gboolean gst_vdec_codec_init(GstVDec * vdec)
{
    VIDDEC2_Params params;
    IVIDDEC2_DynamicParams dynamicParams;
    IVIDDEC2_Status status;
    XDAS_Int32 ret;
    guint8 *ce_name = NULL, *dec_name = NULL, *dec_msg = NULL;

    GST_LOG("BEGIN..");
    params.size = sizeof(VIDDEC2_Params);
    params.maxHeight = 720;
    params.maxWidth = 1280;
    params.maxFrameRate = 0;
    params.maxBitRate = 0;
    params.dataEndianness = XDM_BYTE;
    params.forceChromaFormat = XDM_YUV_422ILE;

    dynamicParams.size = sizeof(IVIDDEC2_DynamicParams);
    dynamicParams.decodeHeader = XDM_DECODE_AU;
    dynamicParams.displayWidth = XDM_DEFAULT;
    dynamicParams.frameSkipMode = IVIDEO_NO_SKIP;
    status.size = sizeof(IVIDDEC2_Status);

    if (!vdec->engineName)
        ce_name = "GstEngine";
    else
        ce_name = vdec->engineName;

    GST_LOG("Opening Engine: %s", ce_name);
    if ((vdec->ce = Engine_open(ce_name, NULL, NULL)) == NULL) {
        GST_WARNING("open Codec_Engine %s failed", ce_name);
        return FALSE;
    }
    //Engine_setTrace(vdec->ce,"*=01234567");

    switch (vdec->codec) {
    case ST_VIDEO_MPEG4:
        dec_name = "mpeg4dec";
        dec_msg  = "MPEG4";
        break;
    case ST_VIDEO_H264:
        dec_name = "h264mpgdec";
        dec_msg  = "H.264";
        break;
    default:
        return FALSE;
        break;
    }

    vdec->decoder = VIDDEC2_create(vdec->ce, dec_name, &params);
    if (vdec->decoder == NULL) {
        GST_ERROR("Error in creating %s decoder", dec_msg);
        return FALSE;
    }
    GST_LOG("Created %s decoder", dec_msg);

    ret = VIDDEC2_control((VIDDEC2_Handle) vdec->decoder,
                       XDM_RESET, &dynamicParams, &status);
    if (ret != VIDDEC2_EOK) {
        GST_WARNING("VIDDEC_control() failed on XDM_RESET");
        return FALSE;
    }
    ret = VIDDEC2_control((VIDDEC2_Handle) vdec->decoder,
                       XDM_SETPARAMS, &dynamicParams, &status);
    if (ret != VIDDEC2_EOK) {
        GST_WARNING("VIDDEC_control() failed on XDM_SETPARAMS");
        return FALSE;
    }
    ret = VIDDEC2_control((VIDDEC2_Handle) vdec->decoder,
                       XDM_GETBUFINFO, &dynamicParams, &status);
    if (ret != VIDDEC2_EOK) {
        GST_WARNING("VIDDEC_control() failed on XDM_GETBUFINFO");
        return FALSE;
    }
    vdec->inBufSize = status.bufInfo.minInBufSize[0];

    return TRUE;
}



static gboolean gst_vdec_sink_event(GstPad * pad, GstEvent * event)
{
    GstVDec *vdec = GST_VDEC(GST_PAD_PARENT(pad));
    gboolean ret = TRUE;

    GST_DEBUG("Got %s event on sink pad", GST_EVENT_TYPE_NAME(event));
    switch (GST_EVENT_TYPE(event)) {
    case GST_EVENT_NEWSEGMENT:
        ret = gst_pad_event_default(pad, event);
        break;
    case GST_EVENT_FLUSH_START:
        ret = gst_pad_event_default(pad, event);
        break;
    case GST_EVENT_FLUSH_STOP:
        ret = gst_pad_event_default(pad, event);
        break;
    case GST_EVENT_EOS:
        ret = gst_pad_push_event(vdec->srcpad, event);
//        ret = gst_pad_event_default(pad, event);
        break;
    default:
        ret = gst_pad_event_default(pad, event);
        break;
    }
    return ret;
}

static gboolean gst_vdec_src_event(GstPad * pad, GstEvent * event)
{
    gboolean res;
    GstVDec *vdec = GST_VDEC(GST_PAD_PARENT(pad));

    GST_LOG("BEGIN..");
    if (vdec->decoder == NULL)
        goto no_decoder;
    res = gst_pad_push_event(vdec->sinkpad, event);
    return res;

no_decoder: {
        GST_DEBUG_OBJECT(vdec, "no decoder, cannot handle event");
        gst_event_unref(event);
        return FALSE;
    }
    GST_LOG("END..");
}
void gst_vdec_reset(GstVDec * vdec)
{
    GValue bytesConsumed = { 0, };

    GST_LOG("BEGIN..");

    vdec->firstFrame = TRUE;
    vdec->srcbuf = NULL;
    vdec->width = 0;
    vdec->height = 0;
    if (vdec->sinkbuf) {
        gst_buffer_unref(vdec->sinkbuf);
        vdec->sinkbuf = NULL;
    }
    vdec->isEof = FALSE;//Who is setting this ? where ?
    vdec->isCodecPropSet = FALSE;
//    vdec->pool_mutex = g_mutex_new();
//    vdec->freesize_mutex = g_mutex_new();
//    vdec->pool_cond = g_cond_new();
//    g_value_set_uint64(&bytesConsumed, vdec->bytesConsumed);
    vdec->bufPoolBase = NULL;
//        Memory_contigAlloc(BUF_POOL_SIZE, Memory_DEFAULTALIGNMENT);
    vdec->poolFreeSize = BUF_POOL_SIZE;
    vdec->poolNextFreeOffset = 0;
    vdec->poolSizeq = g_queue_new();
}

void gst_vdec_cleanup(GstVDec * vdec)
{
    GST_LOG("BEGIN..");
    if (vdec->decoder)
        VIDDEC2_delete((VIDDEC2_Handle) vdec->decoder);
    if (vdec->ce)
        Engine_close(vdec->ce);
//    g_mutex_free(vdec->pool_mutex);
//    g_cond_free(vdec->pool_cond);

    if (vdec->bufPoolBase != NULL)
        Memory_contigFree( vdec->bufPoolBase,
                          BUF_POOL_SIZE);

    GST_LOG("END..");
}

static GstStateChangeReturn
gst_vdec_change_state(GstElement * element, GstStateChange transition)
{
    GstStateChangeReturn result;
    GstVDec *vdec = GST_VDEC(element);

    GST_LOG("BEGIN..");
    switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:
        CERuntime_init();
        gst_vdec_reset(vdec);
        GST_DEBUG_OBJECT(vdec, "State Changed from NULL_TO_READY");
        break;
    case GST_STATE_CHANGE_READY_TO_PAUSED:
        if (!gst_vdec_codec_init(vdec)) {
            GST_ELEMENT_ERROR(vdec, RESOURCE, WRITE, (NULL),
                              ("Cannot initialize codec"));
            return GST_STATE_CHANGE_FAILURE;
        }
        GST_DEBUG_OBJECT(vdec, "State Changed from READY_TO_PAUSED");
        break;
    default:
        break;
    }

    result = GST_ELEMENT_CLASS(parent_class)->change_state(element, transition);
    switch (transition) {
    case GST_STATE_CHANGE_PAUSED_TO_READY:
        gst_vdec_cleanup(vdec);
        GST_DEBUG_OBJECT(vdec, "State Changed from PAUSED_TO_READY");
        break;
    case GST_STATE_CHANGE_READY_TO_NULL:
        CERuntime_exit();
        GST_DEBUG_OBJECT(vdec, "State Changed from READY_TO_NULL");
        break;
    default:
        break;
    }
    GST_LOG("END..");
    return result;
}
